;;; nnjsonfeed.el --- JSON Feed backend for Gnus -*- lexical-binding: t -*-

;; Copyright (C) 2023 Free Software Foundation, Inc.

;; Author: Daniel Semyonov <daniel@dsemy.com>

;; This file is part of GNU Emacs.

;; nnjsonfeed is free software: you can redistribute it and/or modify
;; it under the terms of the GNU General Public License as published by
;; the Free Software Foundation, either version 3 of the License, or
;; (at your option) any later version.

;; nnjsonfeed is distributed in the hope that it will be useful,
;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;; GNU General Public License for more details.

;; You should have received a copy of the GNU General Public License
;; along with nnjsonfeed.  If not, see <https://www.gnu.org/licenses/>.

;;; Commentary:

;; Gnus backend for JSON feeds.

;;; Code:

(require 'nnatom)

(nnoo-declare nnjsonfeed
  nnatom)

(nnatom-define-basic-backend-interface nnjsonfeed)

;;;; JSON feed parser:

(defun nnjsonfeed--read-feed (feed _)
  (with-temp-buffer
    (condition-case e
        (if (file-readable-p feed)
            (insert-file-contents feed)
          (mm-url-insert-file-contents
           (concat "https://" feed)))
      (file-error (nnheader-report nnjsonfeed-backend (cdr e)))
      (:success (or (when-let ((j (json-parse-buffer))) `(,j . 0))
                    (nnheader-report nnjsonfeed-backend "Invalid feed: %s"
                                     feed))))))

(defun nnjsonfeed--read-item (feed _)
  (when-let ((f (car feed)) (i (cdr feed))
             (items (gethash "items" f))
             (item t))
    (unless (length= items i)
      (setcdr feed (1+ i))
      (setq item (aref items i))
      (aset items i nil)
      `(,item . ,feed))))

(defun nnjsonfeed--read-title (feed-or-item)
  (setq feed-or-item (if (listp feed-or-item)
                         (car feed-or-item) feed-or-item))
  (gethash "title" feed-or-item))

(defun nnjsonfeed--read-description (feed)
  (gethash "description" (car feed)))

(defsubst nnjsonfeed--read-author-name (author)
  (when (stringp (setq author (gethash "name" author)))
    (concat author ", ")))

(defun nnjsonfeed--read-authors (feed-or-item)
  (setq feed-or-item (if (listp feed-or-item)
                         (car feed-or-item) feed-or-item))
  (string-trim-right
   (if-let ((authors (gethash "authors" feed-or-item)))
       (mapconcat #'nnjsonfeed--read-author-name authors)
     (if-let ((author (gethash "author" feed-or-item)))
         (nnjsonfeed--read-author-name author)
       ""))
   ", "))

(defun nnjsonfeed--read-id (item)
  (gethash "id" item))

(defun nnjsonfeed--read-date (item mod)
  (when-let ((d (gethash (if mod "date_modified" "date_published") item)))
    (date-to-time d)))

(defun nnjsonfeed--read-links (item)
  (let ((atts 0) links)
    (when-let ((l (gethash "url" item)))
      (push (vector l "Link:" "[Link]") links))
    (when-let ((l (gethash "external_url" item)))
      (push (vector l "Related:" "[Related]") links))
    (append
     links
     (mapcar
      (lambda (a)
        (let* ((l (gethash "url" a))
               (type (gethash "mime_type" a))
               (size (gethash "size_in_bytes" a))
               (dur (gethash "duration_in_seconds" a))
               (title (concat (or (gethash "title" a)
                                  (format "Attachment%s"
                                          (if (< 1 (cl-incf atts))
                                              (format " %d" atts)
                                            "")))
                              (if dur " [" " ")
                              (and dur (format-seconds
                                        "%y year, %d days, %.2h:%z%.2m:%.2s]"
                                        dur))
                              " (" type (and size ", ")
                              (and size (file-size-human-readable size)) ")")))
          (vector l (concat title ":") (concat "[" title "]"))))
      (gethash "attachments" item)))))

(defun nnjsonfeed--read-parts (item)
  (let ((summary (gethash "summary" item))
        (contentt (gethash "content_text" item))
        (contenth (gethash "content_html" item))
        parts)
    (if (not (or summary contentt contenth))
        '((nil ("Content-Type" . "text/html") html links))
      (setq parts nil)
      (and contenth (push (nnatom--read-part contenth "html" t) parts))
      (and summary (push (nnatom--read-part summary "plain"
                                            (not (or contenth contentt)))
                         parts))
      (and contentt (push (nnatom--read-part summary "plain" (not contenth))
                          parts))
      parts)))

(defvoo nnjsonfeed-read-feed-function #'nnjsonfeed--read-feed
  nil nnatom-read-feed-function)
(defvoo nnjsonfeed-read-item-function #'nnjsonfeed--read-item
  nil nnatom-read-article-function)
(defvoo nnjsonfeed-read-title-function #'nnjsonfeed--read-title
  nil nnatom-read-title-function nnatom-read-subject-function)
(defvoo nnjsonfeed-read-description-function #'nnjsonfeed--read-description
  nil nnatom-read-description-function)
(defvoo nnjsonfeed-read-authors-function #'nnjsonfeed--read-authors
  nil nnatom-read-group-author-function nnatom-read-author-function)
(defvoo nnjsonfeed-read-id-function #'nnjsonfeed--read-id
  nil nnatom-read-id-function)
(defvoo nnjsonfeed-read-publish-date-function
    (lambda (item) (nnjsonfeed--read-date item nil))
  nil nnatom-read-publish-date-function)
(defvoo nnjsonfeed-read-modified-date-function
    (lambda (item) (nnjsonfeed--read-date item t))
  nil nnatom-read-update-date-function)
(defvoo nnjsonfeed-read-links-function #'nnjsonfeed--read-links
  nil nnatom-read-links-function)
(defvoo nnjsonfeed-read-parts-function #'nnjsonfeed--read-parts
  nil nnatom-read-parts-function)

(gnus-declare-backend (symbol-name nnjsonfeed-backend) 'address)

(provide 'nnjsonfeed)

;;; nnjsonfeed.el ends here
